"""
run_helpers.py
helper functions for all kinds of experiements
September 2022
"""

import json
import logging
import os
from collections import OrderedDict
from typing import List, Tuple
import glob
from pathlib import Path

import numpy as np
import torch
from omegaconf import DictConfig, OmegaConf


def setup(cfg: DictConfig, caller: str) -> (str, logging.Logger):
    device = "cuda" if torch.cuda.is_available() else "cpu"
    torch.backends.cudnn.benchmark = True
    log = logging.getLogger()
    log.info("\n_________________________________________________\n")
    log.info(f"{caller} running.")
    if cfg.exp_str is not None:
        log.info(f"Running experiment with parameters: {cfg.exp_str}")
    if cfg.hyp.random_seed is not None:
        torch.manual_seed(cfg.hyp.random_seed)
    log.info(OmegaConf.to_yaml(cfg))

    return device, log


def get_train_run_ids(train_experiment_path: str, train_model_checkpoint: str = "model_final.pth", train_run_id: str = None, num_train_runs: int = -1) -> List[str]:
    if train_run_id:
        return [train_run_id]
    train_run_ids = [Path(d_dir).parent.name for d_dir in glob.glob(f"{train_experiment_path}/*/{train_model_checkpoint}")]
    if num_train_runs >= 0:
        train_run_ids = np.random.choice(train_run_ids, size=min(num_train_runs, len(train_run_ids)), replace=False)
    return train_run_ids


def save_results(result: OrderedDict, file_prefix: str = None):
    file_name = "result" if not file_prefix else f"{file_prefix}_result"
    with open(os.path.join(f"{file_name}.json"), "w") as fp:
        json.dump(result, fp, indent=4)


# below functions are for helping generating bash files
def flatten_params(params: OrderedDict) -> OrderedDict:
    output = OrderedDict()
    for group in params:
        if type(params[group]) is dict:
            for config in params[group]:
                new_key = f"{group}.{config}"
                output[new_key] = params[group][config]
        else:
            output[group] = params[group]
    return output


def param_string(param: tuple) -> str:
    if param[1] is None:
        output_string = f"{param[0]}=null "
    else:
        output_string = f"{param[0]}={param[1]} "
    return output_string


def experiment_string(param:tuple) -> str:
    return f"{param[0].split('.')[1]}\={param[1]}_"


def recurse_params(params: OrderedDict) -> List[List[Tuple[str, str]]]:
    output = []
    if len(params.items()) == 0:
        return [[]]

    (key, value) = params.popitem()
    cur = recurse_params(params)
    if not isinstance(value, list):
        print(f"INVALID PARAM: {key}, value: {value}, not a list")

    for v in value:
        e = (key, v)
        for c in cur:
            cp = c.copy()
            cp.append(e)
            output.append(cp)
        if len(output) == 0:
            output.append(e)

    return output
